# distutils: language = c++
# cython: binding = True, language_level=3

include 'version.pxi'
include 'croaring_version.pxi'
include 'abstract_bitmap.pxi'
include 'frozen_bitmap.pxi'
include 'bitmap.pxi'
